import os
import io
import gzip
from typing import Union, Tuple, List, Optional, Dict
from collections import namedtuple, Iterator

from aimrecords.indexing.reader import Reader as IndexReader
from aimrecords.record_storage.consts import (
    RECORD_LEN_SIZE,
    RECORD_OFFSET_SIZE,
    BUCKET_OFFSET_SIZE,
    RECORDS_NUM_SIZE,
    ENDIANNESS,
    RECORDS_NUM,
    BUCKETS_NUM,
    COMPRESSION,
    COMPRESSION_GZIP,
    COMPRESSION_ALGORITHMS,
    DATA_VERSION,
)
from aimrecords.record_storage.utils import (
    read_metadata,
    get_data_fname,
    get_bucket_offsets_fname,
    get_record_offsets_fname,
    current_bucket_fname,
    data_version_compatibility,
)
from aimrecords.indexing.index_key import IndexKey


RecordLocation = namedtuple('RecordLocation', ['bucket_offset',
                                               'bucket_size',
                                               'current_bucket',
                                               'record_local_offset',
                                               ])


class Reader(object):
    def __init__(self, path: str, uncommitted_bucket_visible: bool = False):
        self.path = path
        self.uncommitted_bucket_visible = uncommitted_bucket_visible
        self.metadata = read_metadata(path)
        self.data_version = self.metadata.get('data_version') or DATA_VERSION
        self.validate()

        data_version_compatibility(self.data_version, DATA_VERSION)

        self.compression = self.metadata.get(COMPRESSION)
        assert self.compression in COMPRESSION_ALGORITHMS
        self.buckets_num = self.metadata.get(BUCKETS_NUM) or 0
        self.data_file_bsize = self.metadata.get('data_file_bsize') or [0]
        self.data_records_num = self.metadata.get(RECORDS_NUM) or 0

        self.data_file = open(get_data_fname(self.path), 'rb')
        self._decompressed_bucket_cache = None

        current_bucket_name = current_bucket_fname(self.path)
        if uncommitted_bucket_visible and os.path.exists(current_bucket_name):
            self.current_bucket_file = open(current_bucket_name, 'rb')
            self.curr_bucket_size = os.path.getsize(current_bucket_name)
        else:
            self.current_bucket_file = None
            self.curr_bucket_size = 0

        self.all_records_num = self.data_records_num
        self.record_index_to_offset_list = \
            self._get_record_index_to_offset_list()

        self.indexes = {}  # type: Dict[IndexKey, IndexReader]
        self._iter = None

    def get_records_num(self, indexing: Optional[dict] = None) -> int:
        if indexing is None:
            if self.uncommitted_bucket_visible:
                return self.all_records_num
            else:
                return self.data_records_num
        else:
            if self.uncommitted_bucket_visible:
                return self._get_index(indexing).indexed_records_num()
            else:
                indices_meta = self.metadata.get('indices') or {}
                index_key = IndexKey(indexing)
                index_meta = indices_meta.get(str(index_key)) or {}
                return index_meta.get('indexed_records_num') or 0

    def get(self, index: int,
            indexing: Optional[dict] = None) -> bytes:
        if indexing is not None:
            index = self._get_index(indexing).get(index)

        if index >= self.get_records_num():
            raise IndexError('index out of range')

        record_loc = self.record_index_to_offset_list[index]

        if index >= self.data_records_num:
            record_offset = record_loc.record_local_offset
            self.current_bucket_file.seek(record_offset)
            record_source_fobj = self.current_bucket_file
        elif self.compression is None:
            record_offset = record_loc.record_local_offset
            self.data_file.seek(record_loc.bucket_offset + record_offset)
            record_source_fobj = self.data_file
        else:
            bucket = self._decompressed_bucket(record_loc.bucket_offset,
                                               record_loc.bucket_size)
            bucket.seek(record_loc.record_local_offset)
            record_source_fobj = bucket

        assert record_source_fobj is not None
        size = int.from_bytes(record_source_fobj.read(RECORD_LEN_SIZE),
                              ENDIANNESS)
        record = record_source_fobj.read(size)
        return record

    def close(self):
        self.data_file.close()
        if self.current_bucket_file:
            self.current_bucket_file.close()

        for index in self.indexes.values():
            index.close()

    def get_modification_time(self) -> float:
        if self.uncommitted_bucket_visible:
            current_bucket_name = current_bucket_fname(self.path)
            if os.path.isfile(current_bucket_name):
                return os.path.getmtime(current_bucket_name)

        data_fname = get_data_fname(self.path)
        if os.path.exists(data_fname):
            return os.path.getmtime(data_fname)

        return 0

    def _get_index(self, indexing: dict) -> IndexReader:
        index_key = IndexKey(indexing)
        if index_key not in self.indexes:
            self.indexes[index_key] = IndexReader(self.path, index_key, 'rb')
        return self.indexes[index_key]

    def _decompressed_bucket(self, bucket_offset: int, bucket_size: int):
        if (self._decompressed_bucket_cache is not None
                and self._decompressed_bucket_cache['bucket_offset'] == bucket_offset
                and self._decompressed_bucket_cache['bucket_size'] == bucket_size):
            bucket = self._decompressed_bucket_cache['bucket']
        else:
            self.data_file.seek(bucket_offset)
            compressed_bucket = self.data_file.read(bucket_size)

            if self.compression == COMPRESSION_GZIP:
                bucket = gzip.GzipFile(fileobj=io.BytesIO(compressed_bucket),
                                       mode='rb')
            else:
                raise NotImplementedError('{} decompression is not ' +
                                          'implemented'
                                          .format(self.compression))

            self._decompressed_bucket_cache = {
                'bucket_offset': bucket_offset,
                'bucket_size': bucket_size,
                'bucket': bucket,
            }

        return bucket

    def _get_record_index_to_offset_list(self) -> List[RecordLocation]:
        bucket_offset_records_num_list = []
        with open(get_bucket_offsets_fname(self.path), 'rb') as f_in:
            for _ in range(self.buckets_num):
                bucket_offset = int.from_bytes(f_in.read(BUCKET_OFFSET_SIZE),
                                               ENDIANNESS)
                records_num = int.from_bytes(f_in.read(RECORDS_NUM_SIZE),
                                             ENDIANNESS)

                if len(bucket_offset_records_num_list):
                    prev_bucket_offset = bucket_offset_records_num_list[-1]
                    prev_bucket_offset[1] = (bucket_offset
                                             - prev_bucket_offset[0])
                bucket_offset_records_num_list.append([bucket_offset, 0,
                                                       records_num])

            if len(bucket_offset_records_num_list):
                last_bucket_offset = bucket_offset_records_num_list[-1]
                last_bucket_offset[1] = (self.data_file_bsize[0]
                                         - last_bucket_offset[0])

        record_index_to_offset_list = []
        with open(get_record_offsets_fname(self.path), 'rb') as f_in:
            for record_index in range(self.data_records_num):
                if record_index >= bucket_offset_records_num_list[0][2]:
                    bucket_offset_records_num_list.pop(0)

                bucket_offset = bucket_offset_records_num_list[0][0]
                bucket_size = bucket_offset_records_num_list[0][1]
                record_local_offset = int.from_bytes(
                    f_in.read(RECORD_OFFSET_SIZE),
                    ENDIANNESS)

                record_loc = RecordLocation(bucket_offset, bucket_size,
                                            False, record_local_offset)
                record_index_to_offset_list.append(record_loc)

            if self.uncommitted_bucket_visible:
                self.all_records_num = self.data_records_num
                while True:
                    record_offset_bytes = f_in.read(RECORD_OFFSET_SIZE)
                    if not record_offset_bytes:
                        break

                    record_local_offset = int.from_bytes(record_offset_bytes,
                                                         ENDIANNESS)
                    record_loc = RecordLocation(0, self.curr_bucket_size,
                                                True, record_local_offset)
                    record_index_to_offset_list.append(record_loc)
                    self.all_records_num += 1

        return record_index_to_offset_list

    def validate(self):
        # TODO: implement
        pass


class ReaderIterator(Reader, Iterator):
    def __init__(self, *args, **kwargs):
        super(ReaderIterator, self).__init__(*args, **kwargs)
        self.applied_index = None

    def __getitem__(self, item: Optional[Union[int, Tuple[int, ...], slice]]
                    ) -> iter:
        if item is None:
            item = slice(0, None, 1)

        if isinstance(item, int):
            self._iter = iter((item,))
        elif isinstance(item, tuple):
            if any(map(lambda i: not isinstance(i, int), item)):
                raise TypeError('tuple indices must be integers')
            self._iter = iter(item)
        elif isinstance(item, slice):
            start, stop, step = item.start, item.stop, item.step
            if self.applied_index is None:
                indices_range = range(self.get_records_num())
            else:
                indices_range = range(
                    self._get_index(self.applied_index).indexed_records_num())
            self._iter = iter(indices_range[start:stop:step])
        else:
            raise TypeError('expected slice or tuple of indices')

        return iter(self)

    def __next__(self) -> bytes:
        try:
            idx = next(self._iter)
            return self.get(idx, self.applied_index)
        except (IndexError, StopIteration):
            self._iter = None
            raise StopIteration

    def apply_index(self, index: dict):
        self.applied_index = index



import os
import tempfile
import unittest

from aimrecords.record_storage.writer import Writer
from aimrecords.record_storage.reader import Reader


class TestModTime(unittest.TestCase):
    def test_modification_time(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression=None)
            length = 100
            for index in range(length):
                writer.append_record(str(index).encode())
            writer.close()

            reader = Reader(path)
            first_mod_time = reader.get_modification_time()

            writer = Writer(path, compression=None)
            length = 100
            for index in range(length):
                writer.append_record(str(index).encode())
            writer.close()

            reader = Reader(path)
            second_mod_time = reader.get_modification_time()

            assert first_mod_time != second_mod_time



import os
import tempfile
import unittest

from aimrecords.record_storage.reader import Reader
from aimrecords.record_storage.writer import Writer


class TestBasicStuff(unittest.TestCase):
    def test_simple_int(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression=None)
            length = 1000
            for index in range(length):
                writer.append_record(str(index).encode())
            writer.close()

            reader = Reader(path)
            assert reader.get_records_num() == length

            for index in range(length):
                assert index == int(reader.get(index).decode())

    def test_simple_binary(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression=None)
            length = 5000
            for index in range(length):
                entry = (str(index) * index).encode()
                writer.append_record(entry)
            writer.close()

            reader = Reader(path)
            assert reader.get_records_num() == length

            for index in range(length):
                entry = (str(index) * index).encode()
                assert entry == reader.get(index)



import os
import json
from typing import List

from aimrecords.indexing.index_key import IndexKey
from aimrecords.indexing.utils import list_indices
from aimrecords.record_storage.consts import (
    BUCKET_OFFSET_SIZE,
    RECORDS_NUM_SIZE,
    RECORD_OFFSET_SIZE,
)


def get_data_fname(path: str) -> str:
    return os.path.join(path, 'data_chunk_00000.bin')


def get_record_offsets_fname(path: str) -> str:
    return os.path.join(path, 'record_offsets.bin')


def get_bucket_offsets_fname(path: str) -> str:
    return os.path.join(path, 'bucket_offsets.bin')


def get_metadata_fname(path: str) -> str:
    return os.path.join(path, 'metadata.json')


def current_bucket_fname(path: str) -> str:
    return os.path.join(path, 'current_bucket.bin')


def metadata_exists(path: str) -> bool:
    return os.path.isfile(get_metadata_fname(path))


def write_metadata(path: str, metadata: dict):
    with open(get_metadata_fname(path), 'w') as f_out:
        json.dump(metadata, f_out)


def read_metadata(path: str) -> dict:
    if metadata_exists(path):
        with open(get_metadata_fname(path), 'r') as f_in:
            return json.load(f_in)
    return {}


def data_version_compatibility(prev_version: str, version: str):
    assert 0 <= int(version) - int(prev_version) <= 1


def count_records_num_from_disk(path: str) -> int:
    record_file_path = get_record_offsets_fname(path)
    if not os.path.isfile(record_file_path):
        return 0

    records_num = int(os.path.getsize(record_file_path) / RECORD_OFFSET_SIZE)

    return records_num


def count_buckets_num_from_disk(path: str) -> int:
    bucket_file_path = get_bucket_offsets_fname(path)
    if not os.path.isfile(bucket_file_path):
        return 0

    bucket_info_block_size = BUCKET_OFFSET_SIZE + RECORDS_NUM_SIZE
    buckets_num = os.path.getsize(bucket_file_path) / bucket_info_block_size

    return int(buckets_num)


def get_indices_keys(path: str) -> List[IndexKey]:
    indices_encoded_names_list = list_indices(path)
    indices_keys_list = [IndexKey.from_str(k)
                         for k in indices_encoded_names_list
                         ]
    return indices_keys_list



import tempfile
import unittest

from aimrecords import Storage


class TestMulitpleArtifacts(unittest.TestCase):
    def test_simple_int(self):
        len = 1000

        with tempfile.TemporaryDirectory() as temp_dir:
            storage_writer = Storage(temp_dir, 'w')

            storage_writer.open('loss')
            for i in range(len):
                storage_writer.append_record('loss', str(i).encode())
            storage_writer.close('loss')

            storage_writer.open('accuracy')
            for i in range(len, 2*len):
                storage_writer.append_record('accuracy', str(i).encode())
            storage_writer.close('accuracy')

            del storage_writer

            storage_reader = Storage(temp_dir, 'r')

            storage_reader.open('loss')
            assert storage_reader.get_records_num('loss') == len
            for i, record in enumerate(storage_reader.read_records('loss')):
                assert i == int(record.decode())
            storage_reader.close('loss')

            storage_reader.open('accuracy')
            assert storage_reader.get_records_num('accuracy') == len
            for i, record in enumerate(storage_reader.read_records('accuracy')):
                assert i + len == int(record.decode())
            storage_reader.close('accuracy')

            del storage_reader



import os
from typing import Union, Tuple, Optional
from collections import Iterator

from aimrecords.record_storage.writer import Writer
from aimrecords.record_storage.reader import ReaderIterator
from aimrecords.artifact_storage.consts import (
    STORAGE_DIR_NAME,
)


class Storage:
    WRITING_MODE = 'w'
    READING_MODE = 'r'
    MODES = (
        WRITING_MODE,
        READING_MODE,
    )

    def __init__(self, root_path: str, mode: str):
        self.root_path = root_path
        self.storage_path = self._get_storage_path()
        self._artifacts = {}
        assert mode in self.MODES
        self._mode = mode

        if not os.path.isdir(self.storage_path):
            os.makedirs(self.storage_path)

    def __iter__(self):
        return iter(self._artifacts)

    def open(self, artifact_name: str, *args, **kwargs):
        assert artifact_name not in self._artifacts
        artifact_path = self._get_artifact_path(artifact_name)

        if self._mode == self.WRITING_MODE:
            artifact_instance = Writer(artifact_path, *args, **kwargs)
        elif self._mode == self.READING_MODE:
            artifact_instance = ReaderIterator(artifact_path, *args, **kwargs)
        else:
            raise TypeError('undefined mode')

        self._artifacts.update({
            artifact_name: artifact_instance,
        })

    def append_record(self, artifact_name: str, data: bytes,
                      indexing: Optional[dict] = None) -> int:
        assert self._mode == self.WRITING_MODE

        artifact = self._get_artifact(artifact_name)
        artifact.append_record(data, indexing)

        return artifact.records_num

    def flush(self, artifact_name: str = None):
        assert self._mode == self.WRITING_MODE

        if artifact_name:
            artifact = self._get_artifact(artifact_name)
            artifact.flush()
        else:
            for a in self._artifacts.values():
                a.flush()

    def read_records(self, artifact_name: str,
                     record_indices: Optional[Union[int, Tuple[int, ...],
                                                    slice]] = None,
                     indexing: Optional[dict] = None) -> Iterator:
        assert self._mode == self.READING_MODE

        artifact = self._get_artifact(artifact_name)

        if indexing is not None:
            artifact.apply_index(indexing)

        return artifact[record_indices]

    def get_records_num(self, artifact_name: str,
                        indexing: Optional[dict] = None) -> int:
        artifact = self._get_artifact(artifact_name)
        return artifact.get_records_num(indexing)

    def get_modification_time(self, artifact_name: str) -> float:
        assert self._mode == self.READING_MODE
        artifact = self._get_artifact(artifact_name)
        return artifact.get_modification_time()

    def close(self, artifact_name: str = None):
        if artifact_name:
            artifact = self._get_artifact(artifact_name)
            artifact.close()
            del self._artifacts[artifact_name]
        else:
            while len(self._artifacts):
                artifact_name = list(self._artifacts.keys())[0]
                artifact_inst = self._artifacts[artifact_name]
                artifact_inst.close()
                del self._artifacts[artifact_name]

    def finalize_dangling_artifacts(self):
        assert self._mode == self.WRITING_MODE

        artifact_names = os.listdir(self.storage_path)
        for artifact_name in artifact_names:
            self.open(artifact_name)
            self.close(artifact_name)

    def _get_artifact(self, artifact_name: str
                      ) -> Union[Writer, ReaderIterator]:
        artifact = self._artifacts.get(artifact_name)
        if artifact is None:
            raise ValueError(('artifact {} is not in ' +
                              'storage').format(artifact_name))
        return artifact

    def _get_storage_path(self) -> str:
        return os.path.join(self.root_path, STORAGE_DIR_NAME)

    def _get_artifact_path(self, name: str) -> str:
        return os.path.join(self.storage_path, name)



import tempfile
import unittest

from aimrecords.artifact_storage.storage import Storage


class TestFinalize(unittest.TestCase):
    def test_storage_uncommitted_data_int(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            storage_writer = Storage(temp_dir, 'w')
            storage_writer.open('loss')

            for i in range(10):
                storage_writer.append_record('loss', str(i*10).encode(), {})
            for i in range(10):
                storage_writer.append_record('loss', str(i*20).encode(),
                                             {'subset:': 'train'})
            for i in range(10):
                storage_writer.append_record('loss', str(i*30).encode(),
                                             {'subset': 'val'})
            for i in range(10):
                storage_writer.append_record('loss', str(i*40).encode(),
                                             {'subset:': 'train'})
            for i in range(10):
                storage_writer.append_record('loss', str(i*50).encode(),
                                             {'subset': 'val'})
            storage_writer.flush()

            storage_reader = Storage(temp_dir, 'r')
            storage_reader.open('loss', uncommitted_bucket_visible=False)

            assert storage_reader.get_records_num('loss',
                                                  {'subset:': 'train'}) == 0
            assert storage_reader.get_records_num('loss',
                                                  {'subset': 'val'}) == 0

            storage_reader.close()

            storage_writer_2 = Storage(temp_dir, 'w')
            storage_writer_2.finalize_dangling_artifacts()

            storage_reader = Storage(temp_dir, 'r')
            storage_reader.open('loss', uncommitted_bucket_visible=False)

            assert storage_reader.get_records_num('loss',
                                                  {'subset:': 'train'}) == 20
            assert storage_reader.get_records_num('loss',
                                                  {'subset': 'val'}) == 20
            assert storage_reader.get_records_num('loss', {}) == 10
            assert storage_reader.get_records_num('loss') == 50

            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(0, 10), {})):
                assert i * 10 == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(0, 10),
                                                {'subset:': 'train'})):
                assert i * 20 == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(10, 20),
                                                {'subset:': 'train'})):
                assert i * 40 == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(0, 10),
                                                {'subset': 'val'})):
                assert i * 30 == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(10, 20),
                                                {'subset': 'val'})):
                assert i * 50 == int(record.decode())



import os

from aimrecords.indexing.index_key import IndexKey
from aimrecords.indexing.utils import (
    get_index_offsets_fname,
    get_indices_dir_path,
)
from aimrecords.indexing.consts import (
    INDEX_OFFSET_SIZE,
)


class IndexBase(object):
    def __init__(self, root_path: str, index_key: IndexKey, mode: str):
        self.root_path = root_path
        self.index_key = index_key
        self.name = str(index_key)
        self.num_appended = 0
        self.mode = mode
        self._indices_path = get_indices_dir_path(root_path)
        self._path = os.path.join(self._indices_path, self.name)
        self._offset_file_path = get_index_offsets_fname(self._path)
        self._offset_file = None

    def indexed_records_num(self):
        return int(os.path.getsize(self._offset_file_path) / INDEX_OFFSET_SIZE)

    def close(self):
        self._get_offset_file().close()

    def _get_offset_file(self):
        if not os.path.isdir(self._path):
            if self.mode == 'rb':
                raise FileNotFoundError('index file is not found')
            else:
                os.makedirs(self._path, exist_ok=True)

        if self._offset_file is None:
            if not os.path.exists(self._offset_file_path):
                if self.mode == 'rb':
                    raise FileNotFoundError('index file is not found')
                else:
                    self._offset_file = open(self._offset_file_path, 'wb')
            else:
                self._offset_file = open(self._offset_file_path, self.mode)

        return self._offset_file



from aimrecords.indexing.base import IndexBase

from aimrecords.indexing.consts import (
    INDEX_OFFSET_SIZE,
    ENDIANNESS,
)


class Reader(IndexBase):
    def __len__(self) -> int:
        return self.indexed_records_num()

    def get(self, record_index: int):
        if record_index >= self.indexed_records_num():
            raise IndexError('index out of range')

        offset_file = self._get_offset_file()
        offset_file.seek(INDEX_OFFSET_SIZE * record_index)

        record_loc = int.from_bytes(offset_file.read(INDEX_OFFSET_SIZE),
                                    ENDIANNESS)

        return record_loc



import os
import tempfile
import unittest

from aimrecords.record_storage.reader import Reader
from aimrecords.record_storage.writer import Writer


class TestBucketCompression(unittest.TestCase):
    def test_gzip_compression(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression='gzip')
            length = 1000
            for i in range(length):
                writer.append_record(str(i).encode())
            writer.close()

            reader = Reader(path)
            assert reader.get_records_num() == length

            for index in range(length):
                assert index == int(reader.get(index).decode())



import os
from typing import List

from aimrecords.indexing.consts import (
    INDEX_OFFSET_SIZE,
    INDEX_FILE_NAME,
)


def get_index_offsets_fname(path: str) -> str:
    return os.path.join(path, INDEX_FILE_NAME)


def get_indices_dir_path(path: str) -> str:
    return os.path.join(path, 'indices')


def main_index_file_exists(path: str) -> bool:
    indices_dir_path = get_indices_dir_path(path)
    main_index_path = os.path.join(indices_dir_path, INDEX_FILE_NAME)
    return os.path.isfile(main_index_path)


def list_indices(path: str) -> List[str]:
    indices_dir_path = get_indices_dir_path(path)
    if not os.path.isdir(indices_dir_path):
        return []
    return list(filter(lambda i: i != INDEX_FILE_NAME,
                       os.listdir(indices_dir_path)))


def count_records_num_from_index(path: str, index_key: str) -> int:
    indices_dir_path = get_indices_dir_path(path)
    index_dir_path = os.path.join(indices_dir_path, index_key)
    index_file_path = get_index_offsets_fname(index_dir_path)

    if not os.path.isfile(index_file_path):
        return 0

    records_num = int(os.path.getsize(index_file_path) / INDEX_OFFSET_SIZE)

    return int(records_num)


def normalize_type(val):
    if isinstance(val, str):
        if val == 'True':
            return True
        elif val == 'False':
            return False
        elif val == 'None':
            return None

        try:
            val = int(val)
        except:
            try:
                val = float(val)
            except:
                pass
    return val



import os
import io
import gzip
from shutil import rmtree
from typing import Optional, Dict

from aimrecords.record_storage.consts import (
    RECORD_OFFSET_SIZE,
    RECORD_LEN_SIZE,
    BUCKET_OFFSET_SIZE,
    RECORDS_NUM_SIZE,
    BUCKET_SIZE_KB,
    ENDIANNESS,
    RECORDS_NUM,
    BUCKETS_NUM,
    COMPRESSION_GZIP,
    COMPRESSION_ALGORITHMS,
    DATA_VERSION,
)
from aimrecords.record_storage.utils import (
    write_metadata,
    read_metadata,
    metadata_exists,
    get_bucket_offsets_fname,
    get_data_fname,
    get_record_offsets_fname,
    current_bucket_fname,
    data_version_compatibility,
    count_records_num_from_disk,
    count_buckets_num_from_disk,
    get_indices_keys,
)
from aimrecords.indexing.index_key import IndexKey
from aimrecords.indexing.writer import IndexWriter
from aimrecords.indexing.utils import (
    count_records_num_from_index,
    main_index_file_exists,
)


class Writer(object):
    @staticmethod
    def validate_artifact_path(path):
        # TODO
        return True

    def __init__(self, path: str,
                 compression: Optional[str] = None,
                 rewrite: bool = False):
        assert self.validate_artifact_path(path)
        assert compression in COMPRESSION_ALGORITHMS

        self.path = path
        self.rewrite = rewrite

        self.indexes = {}  # type: Dict[IndexKey, IndexWriter]

        if self.rewrite and self.exists():
            rmtree(self.path)
            os.makedirs(self.path)
        elif not self.exists():
            os.makedirs(self.path)

        file_open_mode = 'wb' if self.rewrite else 'ab'
        self.file_open_mode = file_open_mode

        if self.rewrite:
            self.data_version = DATA_VERSION
            self.compression = compression
            self.data_chunks_num = 0
            self.buckets_num = 0
            self.records_num = 0
            self.indices_meta = {}
        else:
            if metadata_exists(self.path):
                # Restore information from metadata file
                meta = read_metadata(self.path)

                # Get data chunks number from meta file,
                # because only one data chunk is supported currently
                self.data_chunks_num = meta.get('data_chunks_num')

                # Check data format version compatibility
                self.data_version = meta.get('data_version')
                data_version_compatibility(self.data_version, DATA_VERSION)

                # Check whether compression algorithms match
                self.compression = meta.get('compression')
                if self.compression != compression:
                    raise ValueError('already applied {} compression for ' +
                                     '{} artifact'.format(self.compression,
                                                          self.path))

                # Get last indices meta information
                self.indices_meta = meta.get('indices') or {}
            else:
                self.data_version = DATA_VERSION
                self.compression = compression
                self.data_chunks_num = 0
                self.indices_meta = {}
                self.save_metadata(verbose=False)

            # Restore all other info(records num, buckets num etc) from
            # file system to recover previous state
            self.records_num = count_records_num_from_disk(self.path)
            self.buckets_num = count_buckets_num_from_disk(self.path)

            # Restore indices meta info
            indices_keys = get_indices_keys(self.path)
            for index_key in indices_keys:
                index_key_hash = str(index_key)
                index_records_num = count_records_num_from_index(self.path,
                                                                 index_key_hash)

                self.indices_meta.setdefault(index_key_hash, {})
                self.indices_meta[index_key_hash]['indexed_records_num'] = \
                    index_records_num
                if 'keys' not in self.indices_meta[index_key_hash].keys():
                    self.indices_meta[index_key_hash]['keys'] = index_key.index
            # Restore default index info
            if main_index_file_exists(self.path):
                self.indices_meta[''] = {
                    'indexed_records_num': count_records_num_from_index(
                        self.path, ''),
                    'keys': {},
                }

        (
            self.record_offsets_file,
            self.bucket_offsets_file,
            self.current_data_file,
            self.current_bucket_file,
        ) = (
            open(get_record_offsets_fname(self.path), file_open_mode),
            open(get_bucket_offsets_fname(self.path), file_open_mode),
            open(get_data_fname(self.path), file_open_mode),
            open(current_bucket_fname(self.path), file_open_mode),
        )

    def append_record(self, data: bytes,
                      index: Optional[dict] = None):
        current_record_offset = self.current_bucket_file.tell()
        offset_b = current_record_offset.to_bytes(RECORD_OFFSET_SIZE,
                                                  ENDIANNESS)
        data_len_b = len(data).to_bytes(RECORD_LEN_SIZE, ENDIANNESS)

        self.record_offsets_file.write(offset_b)
        self.current_bucket_file.write(data_len_b)
        self.current_bucket_file.write(data)

        if index is not None:
            self.register_index(index)

        self.records_num += 1

        self.current_bucket_file.flush()
        if self._current_bucket_overflow():
            self._finalize_current_bucket()

    def register_index(self, index: dict):
        index_key = IndexKey(index)
        if index_key not in self.indexes:
            self.indexes[index_key] = IndexWriter(self.path, index_key,
                                                  self.file_open_mode)
            self.save_metadata(verbose=False)

        index_inst = self.indexes[index_key]
        index_inst.register_record(self.records_num)

    def flush(self):
        self.current_bucket_file.flush()
        self.record_offsets_file.flush()

        for index in self.indexes.values():
            index.flush()

    def save_metadata(self, verbose=True):
        metadata = {
            'data_version': self.data_version,
            'compression': self.compression,
            'data_chunks_num': self.data_chunks_num,
            'indices': self.indices_meta,
        }

        if verbose:
            records_meta = {
                BUCKETS_NUM: self.buckets_num,
                RECORDS_NUM: self.records_num,
                'record_offsets_bsize': self.record_offsets_file.tell(),
                'bucket_offsets_bsize': self.bucket_offsets_file.tell(),
                'data_file_bsize': [self.current_data_file.tell()],
            }
            metadata = {**metadata, **records_meta}

        for index_key, index in self.indexes.items():
            metadata['indices'].setdefault(index.name, {})
            idx_meta = metadata['indices'][index.name]
            idx_meta.setdefault('indexed_records_num', 0)
            idx_meta.setdefault('keys', index_key.index)
            if verbose and index.num_appended:
                idx_meta['indexed_records_num'] = index.indexed_records_num()

        write_metadata(self.path, metadata)

    def close(self):
        if self.current_bucket_file.tell() > 0:
            self._finalize_current_bucket()

        self.save_metadata()

        self.record_offsets_file.close()
        self.bucket_offsets_file.close()
        self.current_data_file.close()

        for index in self.indexes.values():
            index.close()

        assert self.current_bucket_file.tell() == 0
        self.current_bucket_file.close()
        os.remove(current_bucket_fname(self.path))

    def exists(self):
        return os.path.isdir(self.path)

    def _current_bucket_overflow(self):
        return self.current_bucket_file.tell() / 1024 > BUCKET_SIZE_KB

    def _finalize_current_bucket(self):
        current_bucket_offset = self.current_data_file.tell()
        offset_b = current_bucket_offset.to_bytes(BUCKET_OFFSET_SIZE,
                                                  ENDIANNESS)
        records_num_b = self.records_num.to_bytes(RECORDS_NUM_SIZE, ENDIANNESS)

        self.bucket_offsets_file.write(offset_b)
        self.bucket_offsets_file.write(records_num_b)

        with open(current_bucket_fname(self.path), 'rb') as f_in:
            # depending on size of current_bucket we may want to read it in
            # chunks depending on compression we need to handle this differently
            bucket_data = f_in.read()

            if self.compression == COMPRESSION_GZIP:
                bucket_comp_obj = io.BytesIO(b'')
                with gzip.GzipFile(fileobj=bucket_comp_obj, mode='wb') \
                        as writer:
                    writer.write(bucket_data)
                bucket_data = bucket_comp_obj.getvalue()

            self.current_data_file.write(bucket_data)

        self.buckets_num += 1
        self.current_data_file.flush()
        self.bucket_offsets_file.flush()
        self.current_bucket_file.truncate(0)
        self.current_bucket_file.seek(0)

        self.save_metadata()



import re
from collections import OrderedDict
from typing import Optional
from base58 import b58encode, b58decode

from aimrecords.indexing.utils import normalize_type


class IndexKey(object):
    @staticmethod
    def validate(index: dict):
        if not isinstance(index, dict):
            raise TypeError('index must be a type of dict')

        for i in index.values():
            if type(i) not in [str, int]:
                TypeError('index field must be a type of str or int')

            if re.search('^[A-z0-9_:]+$', str(i)) is None:
                TypeError('index fields can contain only letters, numbers, ' +
                          ' underscore and `:`')

    @staticmethod
    def from_str(encoded_key) -> 'IndexKey':
        indices_items = b58decode(encoded_key).decode().split('-')
        indices = {}
        for item in indices_items:
            k, _, v = item.partition('=')
            indices.setdefault(k, normalize_type(v))
        return IndexKey(indices, encoded_key)

    def __init__(self, index: dict, str_repr: Optional[str] = None):
        self.validate(index)
        self.index = OrderedDict(sorted(index.items()))
        self.str_repr = str_repr

    def __hash__(self) -> int:
        return hash(tuple(self.index.items()))

    def __eq__(self, other) -> bool:
        return hash(self) == hash(other)

    def __str__(self) -> str:
        return self.to_str()

    def to_str(self) -> str:
        if self.str_repr is not None:
            return self.str_repr
        key = '-'.join(['{}={}'.format(i, j) for i, j in self.index.items()])
        return b58encode(key).decode()

    def get_fields(self):
        return tuple(self.index.items())



from aimrecords.indexing.base import IndexBase
from aimrecords.indexing.consts import (
    INDEX_OFFSET_SIZE,
    ENDIANNESS,
)


class IndexWriter(IndexBase):
    def register_record(self, record_index):
        record_index_b = record_index.to_bytes(INDEX_OFFSET_SIZE, ENDIANNESS)
        self._get_offset_file().write(record_index_b)
        self.num_appended += 1

    def flush(self):
        self._get_offset_file().flush()

    def indexed_records_num(self):
        self.flush()
        return super(IndexWriter, self).indexed_records_num()



import os
import tempfile
import unittest

from aimrecords.record_storage.reader import Reader
from aimrecords.record_storage.writer import Writer


class TestReadWrite(unittest.TestCase):
    def test_dirty_read(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression='gzip')

            length = 1000
            for index in range(length):
                writer.append_record(str(index).encode())
                writer.flush()

                reader = Reader(path, uncommitted_bucket_visible=True)
                assert reader.get_records_num() == index + 1
                assert index == int(reader.get(index).decode())
                reader.close()

            writer.close()

    def test_read_write_on_existing_data(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression='gzip')
            length = 1000
            for index in range(length):
                writer.append_record(str(index).encode())
            writer.close()

            writer = Writer(path, rewrite=False, compression='gzip')
            for index in range(length, 2 * length):
                writer.append_record(str(index).encode())
                writer.flush()

                reader = Reader(path, uncommitted_bucket_visible=True)
                assert reader.get_records_num() == index + 1
                assert index == int(reader.get(index).decode())
                reader.close()

            writer.close()

    def test_uncommitted_read_on_closed(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression='gzip')
            length = 1000
            for index in range(length):
                writer.append_record(str(index).encode())
            writer.close()

            reader = Reader(path, uncommitted_bucket_visible=True)
            assert reader.get_records_num() == length
            for index in range(length):
                assert index == int(reader.get(index).decode())
            reader.close()

    def test_read_write(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression=None)

            length = 1000
            for index in range(length):
                writer.append_record(str(index).encode())
                writer.flush()

                reader = Reader(path, uncommitted_bucket_visible=False)
                assert reader.get_records_num() == 0
                reader.close()

            writer.close()



import os
import tempfile
import unittest

from aimrecords.record_storage.reader import ReaderIterator, Reader
from aimrecords.record_storage.writer import Writer
from aimrecords.artifact_storage.storage import Storage


class TestIndexing(unittest.TestCase):
    def test_simple_int(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression=None)
            for index in range(1000):
                writer.append_record(str(index).encode(),
                                     index={'subset': 'train',
                                            'subtask': 'domain'})
            for index in range(500):
                writer.append_record(str(index).encode(),
                                     index={'subset': 'val',
                                            'subtask': 'domain'})
            writer.close()

            writer = Writer(path, compression=None)
            for index in range(100):
                writer.append_record(str(index).encode(),
                                     index={'subset': 'train',
                                            'subtask': 'domain'})
            for index in range(100):
                writer.append_record(str(index).encode(),
                                     index={'subset': 'val',
                                            'subtask': 'domain'})
            writer.close()

            reader = Reader(path)
            for index in range(1000):
                assert index == int(reader.get(index, {'subset': 'train',
                                                       'subtask': 'domain'}))
            for index in range(1000, 1100):
                assert index - 1000 == int(reader.get(index,
                                                      {'subset': 'train',
                                                       'subtask': 'domain'}))
            for index in range(500):
                assert index == int(reader.get(index, {'subset': 'val',
                                                       'subtask': 'domain'}))
            for index in range(500, 600):
                assert index - 500 == int(reader.get(index,
                                                     {'subset': 'val',
                                                      'subtask': 'domain'}))
            reader.close()

    def test_iterator_int(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            writer = Writer(path, compression=None)
            for index in range(1000):
                writer.append_record(str(index).encode(),
                                     index={'subset': 'train',
                                            'subtask': 'domain'})
            for index in range(500):
                writer.append_record(str(index).encode(),
                                     index={'subset': 'val',
                                            'subtask': 'domain'})
            for index in range(100):
                writer.append_record(str(index).encode(),
                                     index={'subset': 'train',
                                            'subtask': 'domain'})
            for index in range(100):
                writer.append_record(str(index).encode(),
                                     index={'subset': 'val',
                                            'subtask': 'domain'})
            writer.close()

            reader = ReaderIterator(path)

            reader.apply_index({'subset': 'train',
                                'subtask': 'domain'})
            for i, record in enumerate(reader[0:1000]):
                assert i == int(record)
            for i, record in enumerate(reader[1000:1100]):
                assert i == int(record)

            reader.apply_index({'subset': 'val',
                                'subtask': 'domain'})
            for i, record in enumerate(reader[0:500]):
                assert i == int(record)
            for i, record in enumerate(reader[500:600]):
                assert i == int(record)
            for i, record in enumerate(reader[0:600]):
                if i < 500:
                    assert i == int(record)
                else:
                    assert i - 500 == int(record)

            reader.close()

    def test_storage_int(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            storage_writer = Storage(temp_dir, 'w')
            storage_writer.open('loss')

            for i in range(1000):
                storage_writer.append_record('loss', str(i).encode(),
                                             {'subset': 'train'})
            for i in range(100):
                storage_writer.append_record('loss', str(i+100).encode(),
                                             {'subset': 'val'})
            for i in range(100):
                storage_writer.append_record('loss', str(i*5).encode(),
                                             {'subset': 'train'})
            for i in range(100):
                storage_writer.append_record('loss', str(i).encode(),
                                             {'subset': 'val'})

            storage_writer.close()
            del storage_writer

            storage_reader = Storage(temp_dir, 'r')
            storage_reader.open('loss')

            assert storage_reader.get_records_num('loss') == 1300
            assert storage_reader.get_records_num('loss',
                                                  {'subset': 'train'}) == 1100
            assert storage_reader.get_records_num('loss',
                                                  {'subset': 'val'}) == 200

            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(0, 1000),
                                                {'subset': 'train'})):
                assert i == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(1000, 1100),
                                                {'subset': 'train'})):
                assert i * 5 == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(0, 100),
                                                {'subset': 'val'})):
                assert i + 100 == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(100, 200),
                                                {'subset': 'val'})):
                assert i == int(record.decode())

            storage_reader.close()
            del storage_reader

    def test_storage_uncommitted_data_int(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            storage_writer = Storage(temp_dir, 'w')
            storage_writer.open('loss')

            for i in range(10):
                storage_writer.append_record('loss', str(i).encode(),
                                             {'subset:': 'train'})
            for i in range(10):
                storage_writer.append_record('loss', str(i+100).encode(),
                                             {'subset': 'val'})
            for i in range(10):
                storage_writer.append_record('loss', str(i*5).encode(),
                                             {'subset:': 'train'})
            for i in range(10):
                storage_writer.append_record('loss', str(i).encode(),
                                             {'subset': 'val'})
            for i in range(10):
                storage_writer.append_record('loss', str(i*12).encode(), {})
            for i in range(10):
                storage_writer.append_record('loss', str(i*22).encode())
            storage_writer.flush()

            storage_reader = Storage(temp_dir, 'r')
            storage_reader.open('loss', uncommitted_bucket_visible=True)

            assert storage_reader.get_records_num('loss') == 60
            assert storage_reader.get_records_num('loss', {}) == 10
            assert storage_reader.get_records_num('loss',
                                                  {'subset:': 'train'}) == 20
            assert storage_reader.get_records_num('loss',
                                                  {'subset': 'val'}) == 20

            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(0, 10), {})):
                assert i * 12 == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(0, 10),
                                                {'subset:': 'train'})):
                assert i == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(10, 20),
                                                {'subset:': 'train'})):
                assert i * 5 == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(0, 10),
                                                {'subset': 'val'})):
                assert i + 100 == int(record.decode())
            for i, record in enumerate(
                    storage_reader.read_records('loss', slice(10, 20),
                                                {'subset': 'val'})):
                assert i == int(record.decode())

            storage_writer.close()
            del storage_writer

            storage_reader.close()
            del storage_reader

    def test_storage_uncommitted_data_false_int(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            storage_writer = Storage(temp_dir, 'w')
            storage_writer.open('loss')

            for i in range(10):
                storage_writer.append_record('loss', str(i).encode(),
                                             {'subset:': 'train'})
            for i in range(10):
                storage_writer.append_record('loss', str(i*10).encode(),
                                             {'subset': 'val'})

            storage_reader = Storage(temp_dir, 'r')
            storage_reader.open('loss', uncommitted_bucket_visible=False)

            assert storage_reader.get_records_num('loss') == 0
            assert storage_reader.get_records_num('loss',
                                                  {'subset:': 'train'}) == 0
            assert storage_reader.get_records_num('loss',
                                                  {'subset': 'val'}) == 0

            storage_reader.close()
            del storage_reader

            storage_writer.close()
            del storage_writer



DATA_VERSION = '0'
RECORD_OFFSET_SIZE = 4
RECORD_LEN_SIZE = 4
BUCKET_OFFSET_SIZE = 8
RECORDS_NUM_SIZE = 8
BUCKET_SIZE_KB = 1024
ENDIANNESS = 'little'

RECORDS_NUM = 'records_num'
BUCKETS_NUM = 'buckets_num'
COMPRESSION = 'compression'

COMPRESSION_GZIP = 'gzip'
COMPRESSION_ALGORITHMS = (
    None,
    COMPRESSION_GZIP,
)



import sys
import os
import io
from shutil import rmtree
from setuptools import find_packages, setup, Command

from aimrecords.__version__ import __version__


here = os.path.abspath(os.path.dirname(__file__))

# Package meta-data.
NAME = 'aimrecords'
DESCRIPTION = 'A record-oriented data format which utilizes Protocol Buffers'
VERSION = __version__
REQUIRES_PYTHON = '>=3.5.0'

# Get packages
packages = find_packages(exclude=('tests',))

# Get long description from the README file
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        LONG_DESCRIPTION = '\n' + f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = DESCRIPTION

# What packages are required for this module to be executed?
REQUIRED = [
    'base58==2.0.1',
]


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = [
        ('rc', None, 'Tag version as a release candidate'),
    ]

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        self.rc = 0

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system(
            '{0} setup.py sdist bdist_wheel --universal'
            .format(sys.executable))

        # self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        if not self.rc:
            self.status('Pushing git tags…')
            os.system('git tag v{0}'.format(VERSION))
            os.system('git push --tags')

        sys.exit()


# Where the magic happens
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    python_requires=REQUIRES_PYTHON,
    install_requires=REQUIRED,
    packages=packages,
    include_package_data=True,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    cmdclass={
        'upload': UploadCommand
    },
)



import os
import tempfile
import unittest

from aimrecords.record_storage.reader import Reader
from aimrecords.record_storage.writer import Writer


class TestWriteAppendModes(unittest.TestCase):
    def test_append_mode_binary(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            length = 1000
            chunks = 5
            chunk_len = length // chunks

            for chunk in range(chunks):
                writer = Writer(path, rewrite=False)
                for index in range(chunk * chunk_len, (chunk + 1) * chunk_len):
                    entry = str(index).encode()
                    writer.append_record(entry)
                writer.close()

            reader = Reader(path)
            assert reader.get_records_num() == length

            for index in range(length):
                entry = str(index).encode()
                assert entry == reader.get(index)

    def test_write_mode_binary(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, 'loss')

            length = 1000

            writer = Writer(path, rewrite=True)
            for index in range(length // 2):
                writer.append_record(b'0')
            writer.close()

            writer = Writer(path, rewrite=True)
            for index in range(length // 2, length):
                entry = str(index).encode()
                writer.append_record(entry)
            writer.close()

            reader = Reader(path)
            assert reader.get_records_num() == length // 2

            for index in range(length // 2, length):
                entry = str(index).encode()
                assert entry == reader.get(index - length // 2)
